import math
import sys


# block of loss functions
def true_loss(model_out: list[float], true_out: list[float]) -> list[float]:
    return [neuron_out - true_out for neuron_out, true_out in zip(model_out, true_out)]


def num_into_interval(number: float, interval: list) -> float:
    return min(max(number, interval[0]), interval[1])


def softmax(values: list) -> list[float]:
    return list(map(lambda x: (math.e ** x) / sum([math.e ** values[j] for j in range(len(values))]), values))


def sigmoid(x) -> float:
    return 1 / (1 + math.exp(x * -1))


def derivative_sigmoid(x) -> float:
    return x * (1 - x)


def relu(x) -> float:
    return max(0, x)


def derivative_relu(x) -> float:
    return 1 if x >= 0 else 0


def tanh(x) -> float:
    return (2 / (1 + math.exp(x * -2))) - 1


def derivative_tanh(x) -> float:
    return 1 - x ** 2


def softplus(x) -> float:
    return math.log(1 + math.exp(x))


def derivative_softplus(x) -> float:
    return 1 / (1 + math.exp(x * -1))


def linear(x) -> float:
    return x


def derivative_linear(x) -> float:
    return 1


def str_to_function(class_string: str):
    return getattr(sys.modules[__name__], class_string)


def str_to_class(class_string: str):
    return getattr(sys.modules[__name__], class_string)


pairs = {
    sigmoid: derivative_sigmoid,
    relu: derivative_relu,
    tanh: derivative_tanh,
    softplus: derivative_softplus,
    linear: derivative_linear
}



import fergana.structures.neurons as neurons
from fergana.math import functions
from abc import ABC, abstractmethod
import random


class __Layer(ABC):
    @abstractmethod
    def __init__(self, number_of_neurons: int):
        pass

    @abstractmethod
    def get_values(self):
        pass

    @abstractmethod
    def clear_data(self):
        pass

    @abstractmethod
    def __copy__(self):
        pass


class InputLayer(__Layer):
    def __init__(self, number_of_neurons: int):
        self.neurons = [neurons.InputNeuron() for neuron in range(number_of_neurons)]

    def get_values(self):
        return [neuron.value for neuron in self.neurons]

    def clear_data(self):
        for neuron in self.neurons:
            neuron.value = None

    def __copy__(self):
        new_layer = InputLayer(len(self.neurons))
        return new_layer


class HiddenLayer(__Layer):
    def __init__(self, number_of_neurons: int, activation_function=functions.sigmoid):
        self.neurons = [neurons.HiddenNeuron(activation_function) for neuron in range(number_of_neurons)]
        self.linked_layer = None

    def __copy__(self):
        new_layer = HiddenLayer(number_of_neurons=len(self.neurons),
                                activation_function=self.neurons[0].activation_function)
        return new_layer

    def get_values(self):
        self.calculate_layer()
        return [neuron.value for neuron in self.neurons]

    def clear_data(self):
        for neuron in self.neurons:
            neuron.value = None
            neuron.gradient = None

    def set_links(self):
        """
        define links between neighbor layers neurons
        """
        for neuron in self.neurons:
            neuron.links[1] = self.linked_layer.neurons
            neuron.links[0] = [random.uniform(-1, 1) for linked_neuron in neuron.links[1]]
            for linked_neuron in neuron.links[1]:
                linked_neuron.linked_neurons.append(neuron)

    def calculate_layer(self):
        """
        update values of layer neurons
        """
        for neuron in self.neurons:
            neuron.update_value()

    def get_gradients(self) -> list[float]:
        """
        :return: list of layer neurons gradients
        """
        return [neuron.gradient for neuron in self.neurons]

    def get_link_weights(self) -> list[list[float]]:
        """
        :return: list of layer neurons individual weights values
        """
        return [neuron.links[0] for neuron in self.neurons]


class OutputLayer(__Layer):
    def __init__(self, number_of_neurons: int, activation_function=functions.sigmoid):
        self.neurons = [neurons.OutputNeuron(activation_function) for neuron in range(number_of_neurons)]
        self.linked_layer = None

    def __copy__(self):
        new_layer = OutputLayer(number_of_neurons=len(self.neurons),
                                activation_function=self.neurons[0].activation_function)
        return new_layer

    def get_values(self):
        self.calculate_layer()
        return [neuron.value for neuron in self.neurons]

    def clear_data(self):
        for neuron in self.neurons:
            neuron.value = None
            neuron.gradient = None

    def set_links(self):
        """
        define links between neighbor layers neurons
        """
        for neuron in self.neurons:
            neuron.links[1] = self.linked_layer.neurons
            neuron.links[0] = [random.uniform(-1, 1) for linked_neuron in neuron.links[1]]
            for linked_neuron in neuron.links[1]:
                linked_neuron.linked_neurons.append(neuron)

    def calculate_layer(self):
        """
        update values of layer neurons
        """
        for neuron in self.neurons:
            neuron.update_value()

    def get_gradients(self) -> list[float]:
        """
        :return: list of layer neurons gradients
        """
        return [neuron.gradient for neuron in self.neurons]

    def get_link_weights(self) -> list[list[float]]:
        """
        :return: list of layer neurons individual weights values
        """
        return [neuron.links[0] for neuron in self.neurons]



from abc import ABC, abstractmethod
import random
from fergana.math import functions


class __Neuron(ABC):

    def __init__(self):
        self.value = None

    @abstractmethod
    def __copy__(self):
        pass


class InputNeuron(__Neuron):
    def __init__(self):
        super().__init__()
        self.linked_neurons = []

    def __copy__(self):
        new_neuron = InputNeuron()
        return new_neuron


class HiddenNeuron(__Neuron):
    def __init__(self, activation_function):
        super().__init__()
        self.activation_function = activation_function
        self.derivative_function = functions.pairs[activation_function]
        self.links = [[], []]
        self.linked_neurons = []
        self.bias_weight = random.random()
        self.gradient = None

    def __copy__(self):
        new_neuron = HiddenNeuron(activation_function=self.activation_function)
        return new_neuron

    def update_value(self):
        """
        update neuron value
        """
        self.value = self.activation_function(sum(
            [neuron.value * weight if neuron.value is not None else 0 for neuron, weight in
             zip(self.links[1],
                 self.links[0])]) + 1 * self.bias_weight)

    def update_gradient(self, local_error: float):
        self.gradient = local_error * self.derivative_function(self.value)

    def create_link(self, linked_neuron):
        """
        create a link between 2 neurons
        :param linked_neuron: neuron to which a link will be created
        """
        self.links[0].append(random.uniform(0, 1))
        self.links[1].append(linked_neuron)
        linked_neuron.linked_neurons.append(self)


class OutputNeuron(__Neuron):
    def __init__(self, activation_function):
        super().__init__()
        self.activation_function = activation_function
        self.derivative_function = functions.pairs[activation_function]
        self.links = [[], []]
        self.bias_weight = random.random()
        self.gradient = None

    def __copy__(self):
        new_neuron = OutputNeuron(activation_function=self.activation_function)
        return new_neuron

    def update_value(self):
        """
        update neuron value
        """
        self.value = self.activation_function(sum(
            [neuron.value * weight if neuron.value is not None else 0 for neuron, weight in
             zip(self.links[1],
                 self.links[0])]) + 1 * self.bias_weight)

    def create_link(self, linked_neuron):
        """
        create a link between 2 neurons
        :param linked_neuron: neuron to which a link will be created
        """
        self.links[0].append(random.uniform(0, 1))
        self.links[1].append(linked_neuron)
        linked_neuron.linked_neurons.append(self)

    def update_gradient(self, local_error: float):
        self.gradient = local_error * self.derivative_function(self.value)



import random


def weight_change(weight, linked_neuron, learning_rate, neuron):
    return weight - learning_rate * neuron.gradient * linked_neuron.value


def gradient_stepping(model, train_data: list[list[float], list[float]], learning_rate: float, loss_func):
    sample_train_data = train_data
    model.clear_values()
    model.data_load(sample_train_data[0])
    model.calculate()
    local_errors = loss_func(model.neural_network[-1].get_values(), sample_train_data[1])
    for neuron in model.neural_network[-1].neurons:  # train output_neurons
        local_error = local_errors[model.neural_network[-1].neurons.index(neuron)]  # optimize that
        neuron.update_gradient(local_error)
        neuron.links[0] = [weight_change(weight, linked_neuron, learning_rate, neuron) for weight, linked_neuron in
                           zip(neuron.links[0], neuron.links[1])]
        neuron.bias_weight = neuron.bias_weight - learning_rate * neuron.gradient
    for layer in model.neural_network[-2:-len(model.neural_network):-1]:
        for neuron in layer.neurons:
            local_error = sum(
                [linked_neuron.links[0][linked_neuron.links[1].index(
                    neuron)] * linked_neuron.gradient if linked_neuron.gradient is not None else 0 for
                 linked_neuron in
                 neuron.linked_neurons])
            neuron.update_gradient(local_error)
            neuron.links[0] = [weight_change(weight, linked_neuron, learning_rate, neuron) for weight, linked_neuron in
                               zip(neuron.links[0], neuron.links[1])]
            neuron.bias_weight = neuron.bias_weight - learning_rate * neuron.gradient


def sgd_optimize(model, train_data: list[list[list[float], list[float]]], learning_rate: float, epochs: int,
                 loss_func):
    for epoch in range(epochs):
        gradient_stepping(model, random.choice(train_data), learning_rate, loss_func)


def minibatch_optimize(model, train_data: list[list[float], list[float]], learning_rate: float,
                       batch_size: int,
                       epochs: int,
                       loss_func):
    for epoch in range(epochs):
        batch = random.choices(train_data, k=batch_size)
        gradient_map = [[[] for gradient in layer.neurons] for layer in model.neural_network]
        values_map = [[[] for value in layer.neurons] for layer in model.neural_network]
        for data in batch:
            new_model = model.__copy__()
            gradient_stepping(new_model, data, learning_rate, loss_func)
            local_gradients = [new_layer.get_gradients() for new_layer in new_model.neural_network]
            local_values = [new_layer.get_values() for new_layer in new_model.neural_network]
            for layer_index, layer_gradients in enumerate(local_gradients):
                for neuron_index, neuron_gradient in enumerate(layer_gradients):
                    gradient_map[layer_index][neuron_index].append(neuron_gradient)
            for layer_index, layer_values in enumerate(local_values):
                for neuron_index, neuron_value in enumerate(layer_values):
                    values_map[layer_index][neuron_index].append(neuron_value)
        for neuron_index, neuron in enumerate(model.neural_network[0].neurons):
            value_list = values_map[0][neuron_index]
            neuron.value = sum(value_list) / len(value_list)
        for layer_index, layer in enumerate(model.neural_network[1:]):
            for neuron_index, neuron in enumerate(layer.neurons):
                value_list = values_map[layer_index + 1][neuron_index]
                gradient_list = gradient_map[layer_index + 1][neuron_index]
                neuron.gradient = sum(gradient_list) / len(gradient_list)
                neuron.value = sum(value_list) / len(value_list)
                neuron.links[0] = [weight_change(weight, linked_neuron, learning_rate, neuron) for weight, linked_neuron
                                   in
                                   zip(neuron.links[0], neuron.links[1])]
                neuron.bias_weight = neuron.bias_weight - learning_rate * neuron.gradient


def batch_optimize(model, train_data: list[list[list[float], list[float]]], learning_rate: float, epochs: int,
                   loss_func):
    for epoch in range(epochs):
        batch = random.choices(train_data, k=len(train_data))
        gradient_map = [[[] for gradient in layer.neurons] for layer in model.neural_network]
        values_map = [[[] for value in layer.neurons] for layer in model.neural_network]
        for data in batch:
            new_model = model.__copy__()
            gradient_stepping(new_model, data, learning_rate, loss_func)
            local_gradients = [new_layer.get_gradients() for new_layer in new_model.neural_network[1:]]
            local_values = [new_layer.get_values() for new_layer in new_model.neural_network]
            for layer_index, layer_gradients in enumerate(local_gradients):
                for neuron_index, neuron_gradient in enumerate(layer_gradients):
                    gradient_map[layer_index + 1][neuron_index].append(neuron_gradient)
            for layer_index, layer_values in enumerate(local_values):
                for neuron_index, neuron_value in enumerate(layer_values):
                    values_map[layer_index][neuron_index].append(neuron_value)
        for neuron_index, neuron in enumerate(model.neural_network[0].neurons):
            value_list = values_map[0][neuron_index]
            neuron.value = sum(value_list) / len(value_list)
        for layer_index, layer in enumerate(model.neural_network[1:]):
            for neuron_index, neuron in enumerate(layer.neurons):
                value_list = values_map[layer_index + 1][neuron_index]
                gradient_list = gradient_map[layer_index + 1][neuron_index]
                neuron.gradient = sum(gradient_list) / len(gradient_list)
                neuron.value = sum(value_list) / len(value_list)
                neuron.links[0] = [weight_change(weight, linked_neuron, learning_rate, neuron) for weight, linked_neuron
                                   in
                                   zip(neuron.links[0], neuron.links[1])]
                neuron.bias_weight = neuron.bias_weight - learning_rate * neuron.gradient


def momentum():
    pass


optimizers = {
    "sgd": sgd_optimize,
    "minibatch": minibatch_optimize,
    "batch": batch_optimize
}



from abc import ABC, abstractmethod
from fergana.math import functions
import json


class __Model(ABC):
    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def __copy__(self):
        """
        Make copy of model, with all same parameters
        """
        pass

    @abstractmethod
    def compile(self):
        """
        Define all links between neurons and make layer structure immutable
        """
        pass

    @abstractmethod
    def clear_values(self):
        """
        clear all values in model, like gradients of neurons, values of neurons and so on
        """
        pass

    @abstractmethod
    def data_load(self, data: list, number_of_layer=0):
        """
        load values of neurons to some layer of model, by default for 0 layer
        :param data: values which will be loaded
        :param number_of_layer: index of layer which will accept data
        """
        pass

    @abstractmethod
    def accuracy_estimate(self, estimate_set: list[list[list, list]], loss_function=functions.true_loss):
        pass

    @abstractmethod
    def save_model(self, name_of_model: str):
        pass

    @staticmethod
    def load_model(file_source: str):
        pass


class PerceptronModel(__Model):
    def __init__(self, *layers):
        self.neural_network = list(layers)

    def __copy__(self):
        new_model = PerceptronModel(
            *[layer.__copy__() for layer in
              self.neural_network])
        new_model.compile()
        for new_layer, layer in zip(new_model.neural_network[1:], self.neural_network[1:]):
            for new_neuron, neuron in zip(new_layer.neurons, layer.neurons):
                new_neuron.links[0] = neuron.links[0]
                new_neuron.bias_weight = neuron.bias_weight
        return new_model

    def compile(self):
        """
        Define all links between neurons and make layer structure immutable
        """
        linked_layer = self.neural_network[0]
        for layer in self.neural_network[1:]:
            layer.linked_layer = linked_layer
            layer.set_links()
            linked_layer = layer
        self.neural_network = tuple(self.neural_network)

    def clear_values(self):
        """
        clear all neurons values, like a gradient and value
        """
        for layer in self.neural_network:
            layer.clear_data()

    def calculate(self) -> list[float]:
        """
        calculate all model layers by getting values of each neuron
        :return: values of output neurons
        """
        for layer in self.neural_network:
            layer.get_values()
        return self.neural_network[-1].get_values()

    def data_load(self, data: list, number_of_layer=0):
        """
        upload values of neurons to some layer of model (by default to 0 layer)
        :param data: uploaded values
        :param number_of_layer: number of layer wherein will be uploaded values
        """
        for neuron, data_element in zip(self.neural_network[number_of_layer].neurons, data):
            neuron.value = data_element

    def accuracy_estimate(self, estimate_set: list[list[list[float], list[float]]], loss_function=functions.true_loss):

        # TODO: Make that work in all cases
        """

                :param estimate_set: data which will be used for determination of accuracy
                :param loss_function: loss function which determines the accuracy of model
                :return: accuracy of model
                """
        accuracy = []
        for data_part in estimate_set:
            self.clear_values()
            self.data_load(data_part[0])
            accuracy_part = loss_function(self.calculate(), data_part[1])
            accuracy.append(sum(accuracy_part) / len(accuracy_part))
        return 1 - sum(map(lambda x: abs(x), accuracy)) / len(accuracy)

    def save_model(self, name_of_model: str):
        # TODO: I think i can save neuron activation function much better
        """
            save a compiled model to a JSON file
            :param name_of_model: name of model, which will be the file name
            """
        model_description = {
            "name": name_of_model,
            "num_of_layers": len(self.neural_network),
            "type_of_model": str(self.__class__.__name__)
        }
        layers = dict()
        layers = {layer_index + 1: [len(layer.neurons), str(layer.__class__.__name__),
                                    [[str(neuron.__class__.__name__), str(neuron.activation_function.__name__),
                                      neuron.bias_weight, neuron.links[0]]
                                     for
                                     neuron in layer.neurons]] for
                  layer_index, layer in
                  enumerate(self.neural_network[1:])}
        layers[0] = [len(self.neural_network[0].neurons), str(self.neural_network[0].__class__.__name__)]
        model_description["layers"] = layers
        with open(name_of_model + ".json", "w") as model_file:
            model_description = json.dumps(model_description)
            model_file.write(model_description)

    @staticmethod
    def load_model(file_source: str):
        new_model = PerceptronModel()
        with open(file_source, "r") as model_file:
            parameters = json.loads(model_file.read())
            layer_specs = parameters["layers"]["0"]
            new_model.neural_network.append(functions.str_to_class(layer_specs[1])(layer_specs[0]))
            for layer_index in range(1, parameters["num_of_layers"]):
                layer_specs = parameters["layers"][str(layer_index)]
                new_model.neural_network.append(
                    functions.str_to_class(layer_specs[1])(layer_specs[0]))
            new_model.compile()
            for layer_index in range(1, parameters["num_of_layers"]):
                layer_specs = parameters["layers"][str(layer_index)]
                for neuron_index, neuron in enumerate(new_model.neural_network[layer_index].neurons):
                    neuron.activation_function = functions.str_to_function(layer_specs[2][neuron_index][1])
                    neuron.derivative_function = functions.pairs[
                        functions.str_to_function(layer_specs[2][neuron_index][1])]
                    neuron.bias_weight = layer_specs[2][neuron_index][2]
                    neuron.links[0] = layer_specs[2][neuron_index][3]
            return new_model


class ConvolutionalModel(__Model):
    pass
